/* vi:ts=4:sw=4
 *
 * VIM - Vi IMitation Kanji extension
 *
 * Code Contributions By:	Atsushi Nakamura		ann@mrit.mei.co.jp
 */

/*
 * Code to handle kanji conversion.
 */

#include "vim.h"
#include "globals.h"
#include "proto.h"
#include "param.h"
#ifdef KANJI
#include "kanji.h"
#endif

#ifdef KANJI

#ifdef UNIX
core_handle()
{
	char *text;

	settmode(0);
	stoptermcap();
	flushbuf();
	stopscript();

	signal(SIGSEGV, SIG_DFL);
	signal(SIGBUS,  SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGILL,  SIG_DFL);
	signal(SIGLOST, SIG_DFL);

	text = modname(Filename, ".sav");
	printf("\n\nJVIm is going to dead. \n");
	if (!writeit(text, (linenr_t)1, line_count, FALSE, FALSE, TRUE))
		printf("\nI couldn't save the current text.\n");

	kill(getpid(), SIGQUIT);
}
#endif

kanji_init()
{
	KANJI_FCODE = KANJI_NEW;
#ifdef UNIX
	signal(SIGSEGV, core_handle);
	signal(SIGBUS,  core_handle);
	signal(SIGQUIT, core_handle);
	signal(SIGILL,  core_handle);
	signal(SIGLOST, core_handle);
#endif
}

/*
 * convert EUC letter into suitable letter.
 */
	void
kanjito(k1, k2, code)
	u_char	*k1, *k2;
	char	code;
{
	u_char c1, c2;

	switch(code)
	{
	  case KCODE_NJIS:
	  case KCODE_JIS:
		*k1 &= 0x7f;
		*k2 &= 0x7f;
		return;

	  case KCODE_SJIS:
	  	c1 = *k1 & 0x7f;
	  	c2 = *k2 & 0x7f;
	    *k1 = ((c1 - 1) >> 1) + ((c1 <= 0x5e) ? 0x71 : 0xb1);
		*k2 = c2 + ((c1 & 1) ? ((c2 < 0x60) ? 0x1f : 0x20) : 0x7e);
	    return;

	  default:
		return;
	}
}

/*
 * return kanji shift-in string
 */
	u_char *
kanjiin(code)
	char	code;
{
	switch(code)
	{
	  case KCODE_JIS:
		return (u_char *)"\033$@";

	  case KCODE_NJIS:
		return (u_char *)"\033$B";

	  default:
		return (u_char *)"";
	}
}

/*
 * return kanji shift-out string
 */
	u_char *
kanjiout(code)
	char	code;
{
	switch(code)
	{ 
	  case KCODE_NJIS:
		return (u_char *)"\033(B";

	  case KCODE_JIS:
		return (u_char *)"\033(J";

	  default:
		return (u_char *)"";
	}
}

/*
 * convert buffer to EUC code
 */
static int kanji, jis, euc, njis;

	char
kanjiconvfrom(lnum, lend, code)
	linenr_t	lnum, lend;
	char		code;
{
	u_char	*ptr, *new;

	kanji = FALSE;
	jis = euc = njis = 0;

	for(; lnum <= lend; ++lnum)
	{
		ptr = (u_char *)nr2ptr(lnum);
		new = kanjiconvsfrom(ptr, code);
		if (new != ptr) free_line(replaceline(lnum, new));
	}

	if (code == KCODE_ANY)
	{
		char	judge;
		if (!jis && !euc && !njis)
			judge = KCODE_NONE;
		else if (njis)
			judge = KCODE_NJIS;
		else if (jis)
			judge = KCODE_JIS;
		else
			judge = KCODE_EUC;
		return judge;
	}
	else
		return code;
}

	u_char *
kanjiconvsfrom(ptr, code)
	u_char	*ptr;
	char	code;
{
	u_char	*top, *ptr2;
	u_char	c1, c2;

	top = ptr;

	switch(code)
	{
	  case KCODE_EUC:
	  	return top;
	
	  case KCODE_NONE:
		for(;*ptr; ptr++)
			 *ptr |= 0x80;
		return top;

	  case KCODE_SJIS:
		for(;*ptr; ptr++)
			if (IsKanji(c1 = *ptr) && (c2 = *(ptr+1)))
			{	int i;
				i = (c1 << 1) - ((c1 <= 0x9f)? 0x00e1 : 0x0161);
				if (c2 < 0x9f)
				{
					c2 -= ((c2 > 0x7f) ? 0x20 : 0x1f);
				}
				else
				{
					c2 -= 0x7e;
					c1 = i + 1;
				}
				*ptr++ = c1 | 0x80;
				*ptr   = c2 | 0x80;
			}
		return top;

	  case KCODE_NJIS:
	  case KCODE_JIS:
	  case KCODE_ANY:

		ptr2 = ptr;
		while(*ptr)
		{
			if (IsKanji(*ptr)) euc++;

			if (kanji)
			{
				if (*ptr == ESC)
				{
					if (*++ptr == '(')
					{
					 	if (*++ptr == 'J' || *ptr == 'H' || *ptr == 'B' ||
						    code == KCODE_ANY) /* Kanji Out */
						{
							kanji = FALSE;
							ptr++;
							continue;
						}
						else
							ptr-=2;
					}
					else
						ptr--;
				}

				*(ptr2++) = (*ptr++) | 0x80;
			}
			else
			{
				if (*ptr == ESC)
				{
					if (*++ptr == '$')
					{
					 	if (*++ptr == '@' || *ptr == 'B' ||
							code == KCODE_ANY)		/* Kanji In */
						{
							if      (*ptr == '@') jis++;
							else if (*ptr == 'B') njis++;

							kanji = TRUE;
							ptr++;
							continue;
						}
						else
							ptr-=2;
					}
					else
						ptr--;
				}

				*ptr2++ = *ptr++;
			}
		}
		*ptr2 = NUL;
		return top;
	}
	return top;
}

/*
 * convert buffer from EUC code
 */
kanjiconvto(lnum, lend, code)
	linenr_t	lnum, lend;
	char		code;
{
	u_char	*ptr, *new;

	for(; lnum <= lend; ++lnum)
	{
		ptr = (u_char *)nr2ptr(lnum);
		new = kanjiconvsto(ptr, code);
		if (new != ptr) free_line(replaceline(lnum, new));
	}
}

	u_char *
kanjiconvsto(ptr, code)
	u_char	*ptr;
	char	code;
{
	u_char	*top, *ptr2;
	u_char	*kin, *kout, *cp;
	int		nshift, nchar;

	top = ptr;

	switch(code)
	{
	  case KCODE_EUC:
	  	return top;
	
	  case KCODE_NONE:
		for(;*ptr; ptr++)
			 *ptr &= ~0x80;
		return top;

	  case KCODE_SJIS:
		for(;*ptr; ptr++)
			if (IsKanji(*ptr) &&  *(ptr+1))
			{
				kanjito(ptr, ptr + 1, code);
				ptr++;
			}
		return top;

	  case KCODE_NJIS:
	  case KCODE_JIS:
	  case KCODE_ANY:

		kin  = kanjiin(code);
		kout = kanjiout(code);

		kanji = 0;
		for(nshift = 0; *ptr; ptr++)
		{
			if (kanji !=  (*ptr & 0x80))
			{
				nshift++;
				kanji ^= 0x80;
			}
		}
		if (kanji) nshift++;
		nchar = ptr - top;

		if (!nshift) return top;

		ptr = top;
		top = ptr2 = (u_char *)alloc_line(nchar + nshift * 3);

		kanji = 0;
		while(*ptr) 
		{
			if (kanji !=  (*ptr & 0x80))
			{
				if (kanji)
					for(cp = kout; *cp;)
						*ptr2++ = *cp++;
				else
					for(cp = kin; *cp;)
						*ptr2++ = *cp++;
				kanji ^= 0x80;
			}
			*ptr2++ = (*ptr++ & 0x7f);
		}
		if (kanji)
			for(cp = kout; *cp;)
				*ptr2++ = *cp++;
		*ptr2 = 0;
		return top;
	}
	return top;
}

kanji_align()
{
	int col;
	char *line;

	line = pos2ptr(&Curpos);
	for(col = 0; col < Curpos.col; col++, line++)
		if (IsKanji(*line))
		{
			line++;
			col++;
		}
	if (col != Curpos.col) Curpos.col--;
}

#endif /* KANJI */
